/*
 *	Project name:	Multi-CarAudio Monitor
 *	Copyright:		(c)2009 CyberPit HILO
 *
 *	Description:
 * 		This code is created for the multi purpose car audio accessary.
 *		Those features are using on-chip ADC's 10bit conversion
 *		Calculated results are displayed on 8x2 LCD display module
 *		Available features are...
 *			- Battery Voltage Monitor
 *			- Power Amplifire Output Power
 *			- Headroom Margin from Clipping Level
 *			- Stereo Level Bar Display
 *	Display:		LCD (ACM0802C-NLW-BBH)
 *	SW:             PICC-PRO
 *	MCU:			PIC16F690
 *	MCU Configurations:
 *		Oscillator:	HS, 8.0000 MHz, No-Brownout, Internal Reset
 *
 *	Port Assign:
 *				MODE_SW:	 RA3	(IN)
 *				LCD_E:       RB7	(OUT)
 *				LCD_RW:      RB6	(OUT)
 *				LCD_RS:      RB5	(OUT)
 *				LCD_PUSH SW: RB4	(OUT)
 *				LCD_Data:    RC[0..7] (Bidirectional)
 * ADC Assign:
 *				AN0:		Headroom Reference Trim
 *				AN1:		Battery Volatage
 *				AN2:		Audio Level R
 *				AN3:		Audio Level L
 */

#include <pic.h>
#include <htc.h>	// Required to interface with delay routines
#include <stdio.h>
#include <string.h>
#include <math.h>
#ifndef _XTAL_FREQ
 // Unless already defined assume 4MHz system frequency
 // This definition is required to calibrate __delay_us() and __delay_ms()
 #define _XTAL_FREQ 4000000
#endif

/*
====<< PIC Configuration bit definitions >>====
*/
__CONFIG(INTIO & WDTDIS & PWRTDIS & MCLRDIS & UNPROTECT & BORDIS & IESODIS & FCMDIS);

/*
====<< Stored Setting Data Address on EEPROM >>====
*/
#define	EEPROM_SPEAKER_IMP				0x02	// SpeakerImpedance 1 through 16
#define EEPROM_DEFAULT_DISP				0x03	// Startup Dsiplay Mode
#define	EEPROM_BTL_AMP_FLAG				0x04	// Set 2 when BTL, otherwose 1, Undefined 0

/*
----< Defined Values >----
*/
#define	HOLDTIME	10

/*
====<< LCD Control Commands >>====
*/
#define LCD_DISPLAY_CLEAR                0x01
#define LCD_CURSOR_AT_HOME               0x02
#define LCD_DISPLAY_ON_BLINK_CURSOR      0x0f
#define LCD_DISPLAY_ON_NONBLINK_CURSOR   0x0e
#define LCD_DISPLAY_ON_NO_CURSOR         0x0c
#define LCD_DISPLAY_OFF                  0x08
#define LCD_CURSOR_RIGHT                 0x14
#define LCD_CURSOR_LEFT                  0x10
#define LCD_DISPLAY_SHIFT_RIGHT          0x1c
#define LCD_DISPLAY_SHIFT_LEFT           0x18
#define LCD_CG_ADR_SET                   0x20
#define LCD_SET_DATABUS8                 0x38
#define LCD_CURSOR_INC_NOSHIFT           0x06
#define LCD_CURSOR_DEC_NOSHIFT           0x04
#define LCD_CURSOR_INC_SHIFT             0x07
#define LCD_CURSOR_DEC_SHIFT             0x05

/*
====<< Font Data >>===
*/
char character_P[] = {31,17,21,17,23,23,31,0};
char character_H[] = {31,21,21,17,21,21,31,0};
char character_BAT[] = {10,31,31,31,31,31,31,0};
char character_sbar1[] = {16,16,16,16,16,16,16,16};
char character_sbar2[] = {20,20,20,20,20,20,20,20};
char character_sbar3[] = {21,21,21,21,21,21,21,21};
char character_L[] = {31,23,23,23,23,23,17,31};
char character_R[] = {31,19,21,21,19,21,21,31};

const char*     mode_text[4] = 	{
								"BATTERY ", 
								"POWER   ",
								"HEADROOM", 
								"LEVELBAR"
								};
const char	ad_message[]=
	{
	"        by CyberPit HILO        "
	};							

/*
====<< Global Variables on RAM >>====
*/
static unsigned int   	digit, number, hr_ref, hold_timer, hold_value;
static char     		mode, spkimp, amptype;
static char 			sbuf[10];

/*
====<< Interrupt Service Routine >>====
*/
unsigned int   kbmaskcnt;
char     kbdhit,kholdflag,key1,last_key1;

void interrupt timer0_isr(void)
{
last_key1 = key1;
key1 = (0x10 & PORTB);                 //Set if Key1 is not pushed
if(kbmaskcnt)
	{
     --kbmaskcnt;
	if(key1)
		kholdflag = 0;	// Clear Key Hold Flag if key released
	}
else if((last_key1 == key1) && (key1 == 0x00))
    {  // if key1 pushed
	if(kholdflag == 0)
		{
	    kbdhit = 0x01;	// Set Key one push flag
		kholdflag = 1;	// Set Hold Flag
		}
	else{
		++kbdhit;	// Increment Key Holded flag
		}
    kbmaskcnt = 1000;
    }
TMR0   =   0;                          // reset TIMER0 value
INTCON = 0x20;                         // Clear T0IF
}

/*
----< Long Delay Function >----
*/
delay_10ms(unsigned int loop)
{
while(--loop)
__delay_ms(10);                         // Wait for Human Ready
}


/*
----< 10bits Read A/D Converter >----
*/
unsigned int ADC_Read(unsigned char channel)
{
	channel&=0x0F;			// Mask nibbles
	ADCON0&=0xC3;			// Clear Current
	ADCON0|=(channel<<2);	// Set CH
	__delay_ms(1);			// Wait for MUX switching
	GODONE=1;				// Conv. Start
	while(GODONE)continue;	// Wait for complete
	return((((unsigned int)ADRESH<<8) | (unsigned int)ADRESL));		//Return Result
}

 /*
----<  Check LCD Busy Status >----
*/
char acm0802_status_read(void)
{
unsigned char result;
TRISC = 0xff;			// Set PORTC as input
PORTB = 0x40;			// Set Read
PORTB = 0xC0;			// Set E
result = PORTC;
if((0x80 & result))
     result = 0;    	// returns 0 when BF==1
PORTB = 0x00;          	// Clear E
TRISC = 0;				// Set PORTC as output
return(result);
}

 /*
----< LCD Regiter Write >----
*/
void acm0802_reg_write(unsigned char rdata)
{
PORTB = 0x00;
__delay_ms(2);
PORTB = 0x80;
__delay_ms(2);
PORTC = rdata;
__delay_ms(2);
PORTB = 0x00;
__delay_ms(2);
}

/*
----< LCD Data Write >----
*/
void acm0802_data_write(unsigned char rdata)
{
//while (!(0 == acm0802_status_read()))       ; // wait if busy
PORTB = 0x20;
__delay_ms(1);
PORTB = 0xA0;
__delay_ms(1);
PORTC = rdata;
__delay_ms(1);
PORTB = 0x20;
__delay_ms(1);
}

/*
----< LCD Display Address Set >----
*/
void acm0802_DD_AddressSet(unsigned char columm)
{
acm0802_reg_write((0x80 + columm));
}

/*
----< Custom Character Font Data Set >----
*/
void CustomChar(void)
{
char i;
acm0802_reg_write((0x40 + 0));
    for (i = 0; i<=7; i++) acm0802_data_write(character_P[i]);
    for (i = 0; i<=7; i++) acm0802_data_write(character_H[i]);
    for (i = 0; i<=7; i++) acm0802_data_write(character_BAT[i]);
    for (i = 0; i<=7; i++) acm0802_data_write(character_sbar1[i]);
    for (i = 0; i<=7; i++) acm0802_data_write(character_sbar2[i]);
   	for (i = 0; i<=7; i++) acm0802_data_write(character_sbar3[i]);
    for (i = 0; i<=7; i++) acm0802_data_write(character_L[i]);
    for (i = 0; i<=7; i++) acm0802_data_write(character_R[i]);
acm0802_reg_write(LCD_CURSOR_AT_HOME);
}

/*
----< Initialize  LCD Module >----
*/
void acm0802_init(void)
{
__delay_ms(50);
 acm0802_reg_write(LCD_SET_DATABUS8);         /* 8bit Connection*/
__delay_ms(5);
 acm0802_reg_write(LCD_DISPLAY_ON_NO_CURSOR);
__delay_ms(1);                                 // BF check enabled from here
 acm0802_reg_write(LCD_DISPLAY_CLEAR);
__delay_ms(2);                                 // Need to wait 1.65mS
 acm0802_reg_write(LCD_CURSOR_INC_NOSHIFT);
__delay_ms(1);
}

/*
----< Print String on LCD >----
*/
void acm0802_text_write(char *string, char place)
{
char length;
char *cpoint;
cpoint=string;
length=strlen(string);
if(place)	// place==0 means print from current position otherwise as defined line
	acm0802_DD_AddressSet(0x01 + (0x40 * (--place))); /* Left Justified */
while(length > 0)
    {
    if(0 != (acm0802_status_read()))    // Skip if busyflug
         {
         acm0802_data_write(*cpoint);
         ++cpoint;
         --length;
         }
    }
}

/*
---< Display 4 digits >----
*/
void Disp_Digits(unsigned int value, char unit)
{
number=value;
sbuf[0] = ' ';
sbuf[1] = ' ';
sbuf[2] = ' ';
digit = number / 1000u ;             // extract thousands digit
sbuf[3] = digit + 0x30;        // and store it to PORTD array
digit = (number / 100u) % 10u;       // extract hundreds digit
sbuf[4] = digit + 0x30;        // and store it to PORTD array
digit = (number / 10u) % 10u;        // extract tens digit
sbuf[5] = digit + 0x30;        // and store it to PORTD array
digit = number % 10u;                // extract ones digit
sbuf[6] = digit + 0x30;        // and store it to PORTD array
sbuf[7] = unit;
sbuf[8] = 0;
acm0802_text_write(sbuf, 2);
}


/*
----< Display Battry Voltage >----
*/
void Disp_Voltage(unsigned int adc_value)
{
if(hold_value > adc_value)
	{
	hold_value = adc_value;	// Update held Value
	hold_timer = HOLDTIME + HOLDTIME;	// Reset Hold Timer (2Times)
	}
else
	{
	if(hold_timer)
		--hold_timer;
	else
		hold_value = adc_value;
	}
number = 1.541 * hold_value;   // Get 10-bit results of AD conversion
sbuf[0] = 0x0a;
sbuf[1] = 0x20;
digit = number / 1000u ;             // extract thousands digit
sbuf[2] = digit + 0x30;        // and store it to PORTD array
digit = (number / 100u) % 10u;       // extract hundreds digit
sbuf[3] = digit + 0x30;        // and store it to PORTD array
sbuf[4] = '.';
digit = (number / 10u) % 10u;        // extract tens digit
sbuf[5] = digit + 0x30;        // and store it to PORTD array
digit = number % 10u;                // extract ones digit
sbuf[6] = digit + 0x30;        // and store it to PORTD array
sbuf[7] = 'V';
sbuf[8] = 0x00;       
if(sbuf[2] == '0')
	{
	sbuf[2] = ' ';				// Leading Zero Suppress
	}
acm0802_text_write(sbuf, 2);
}

/*
----< Display Headroom >----
*/
void Disp_Headroom(unsigned int adc_value)
{
if(hold_value < adc_value)
	{
	hold_value = adc_value;	// Update Hold Value
	hold_timer = HOLDTIME;	// Reset Hold Timer
	}
else
	{
	if(hold_timer)			// Holds peak if not yet
		--hold_timer;
	else
		hold_value = adc_value;
	}
sbuf[0] =  0x09;
if(hold_value > 27)
	{
	if((unsigned int)hold_value < (unsigned int)hr_ref)
        { 
		sbuf[1] =  '-';
		number = (unsigned int)(-200 * log10((double)(hold_value/(double)hr_ref)));
		}
	else    
		{  
		sbuf[1] =  '+';
		number = (unsigned int)(200 * log10((double)(hold_value/(double)hr_ref)));
		}
	digit = (number / 100u) % 10u;       // extract hundreds digit
	sbuf[2] = digit + 0x30;        // and store it to PORTD array
	digit = (number / 10u) % 10u;        // extract tens digit
	sbuf[3] = digit + 0x30;        // and store it to PORTD array
	sbuf[4] = '.';
	digit = number % 10u;                // extract ones digit
	sbuf[5] = digit + 0x30;        // and store it to PORTD array
	}
else
	{
	sbuf[2]=sbuf[3]=sbuf[4]=sbuf[5]='-';
	}		
sbuf[6] = 'd';
sbuf[7] = 'B';
sbuf[8] = 0x00;
acm0802_text_write(sbuf, 2);
}

/*
----< Display LevelBar >----
*/
void Disp_LevelBar(unsigned int adc_value,char line)
{
char	loop;
int		data;
loop = 0;
if( line == 1 )
	{
	acm0802_DD_AddressSet(0x1); /* Cursor at Left of line */
	acm0802_data_write(0x06);				// Reversed 'L'
	}
else{
	acm0802_DD_AddressSet(0x41); /* Cursor at 2nd Left of line */
	acm0802_data_write(0x07);				// Reversed 'L'
	}	
data =  (int)(-20 * log10(((double)adc_value/(double)hr_ref)))-1;
if(data < 0)
	{ 	// Over Zero means Full Scale
	sbuf[6]=sbuf[5]=sbuf[4]=sbuf[3]=sbuf[2]=sbuf[1]=0x05; 
	sbuf[7] =  0xff;	// Level Overflow Charcter 
	acm0802_text_write(sbuf+1, 0);
	}
else	
	{
	if(data < 18)
		{
		while(loop < ((18 - data)/3))	
			{
			acm0802_data_write( 0x05 );
			++loop;
			}
		if((18 - data) %3)	
			acm0802_data_write( 0x02 + (18 - data)%3);
		}
	while( loop < 8)
		{
		acm0802_data_write( 0x20 );
		++loop;
		}
	}
}

/*
----< Display Peak Power >----
*/
void Disp_Power(unsigned int adc_value)
{
if(hold_value < adc_value)
	{
	hold_value = adc_value;	// Update Hold Value
	hold_timer = HOLDTIME;	// Reset Hold Timer
	}
else
	{
	if(hold_timer)			// Holds peak if not yet
		--hold_timer;
	else
		hold_value = adc_value;
	}
//adc_value = hold_value * amptype;	// Voltage 2times if using BTL amplifire 
number =(( ((2.2261 * hold_value * amptype) - 26) * ((2.2261 * hold_value * amptype) - 26))/ (1000 * spkimp));   // Calc (Vsp^2)/R
sbuf[0] = sbuf[1] = 0x08;
sbuf[1] = sbuf[1] = 0x20;
digit = number / 1000u ;             // extract thousands digit
sbuf[2] = digit + 0x30;        // and store it to PORTD array
digit = (number / 100u) % 10u;       // extract hundreds digit
sbuf[3] = digit + 0x30;        // and store it to PORTD array
digit = (number / 10u) % 10u;        // extract tens digit
sbuf[4] = digit + 0x30;        // and store it to PORTD array
digit = number % 10u;                // extract ones digit
sbuf[5] = '.';
sbuf[6] = digit + 0x30;        // and store it to PORTD array
sbuf[7] = 'W';
sbuf[8] = 0x00;      
if(sbuf[2] == '0')				// Leading Zero Suppress
	{
	sbuf[2] = ' ';
	if(sbuf[3] == '0')
		sbuf[3] = ' ';
	}
acm0802_text_write(sbuf, 2);
}

#ifdef DEBUG
/*
----< Display Ref Voltage >----
*/
void Disp_Ref(unsigned int adc_value)
{
acm0802_DD_AddressSet(0x41); /* Left Justified */
number=amptype * ((2.2261 * adc_value) - 26);	// Calc Compensation
sbuf[0] = 0x09;
sbuf[1] = 0x20;
digit = number / 1000u ;             // extract thousands digit
sbuf[2] = digit + 0x30;        // and store it to PORTD array
digit = (number / 100u) % 10u;       // extract hundreds digit
sbuf[3] = digit + 0x30;        // and store it to PORTD array
sbuf[4] = '.';
digit = (number / 10u) % 10u;        // extract tens digit
sbuf[5] = digit + 0x30;        // and store it to PORTD array
digit = number % 10u;                // extract ones digit
sbuf[6] = digit + 0x30;        // and store it to PORTD array
sbuf[7] = 'V';
sbuf[8] = 0x00;        			// Terminater
if(sbuf[2] == '0')				// Leading Zero Suppress
	sbuf[2] = ' ';
acm0802_text_write(sbuf, 0);
}
#endif

/*
---< Configuration MENU >---
*/
char*	conf_messages[12] = { "SETTING ",	"START   ",	"SPK IMP=",	"        "
							,"Writting",	"DEFAULT=",	"DISPLAY?",	"Finished"
							,"BTL AMP?",	"NO(GND+)",	"YES(+,-)", "-1dB/dot"};


static void configmenu(void)
{
char current_value;
acm0802_text_write(conf_messages[0],1);
acm0802_text_write(conf_messages[1],2);
delay_10ms(150);                         // Wait for Human Ready
acm0802_text_write(conf_messages[2],1);	 // Ask Impedance?
acm0802_text_write(conf_messages[3],2);
delay_10ms(50);                         // Wait for Human Ready
kbdhit=2;
while(kbdhit >=2)
	{
    acm0802_DD_AddressSet(0x41); /* Left Justified */
	current_value =  1+ kbdhit%16;
	Disp_Digits((unsigned int)current_value, 0xf4);
	}
eeprom_write(EEPROM_SPEAKER_IMP,current_value);
acm0802_text_write(conf_messages[4],1);
delay_10ms(200);                         // Wait for Human Ready

acm0802_text_write(conf_messages[8],1);	 // Ask BTL?
acm0802_text_write(conf_messages[3],2);
delay_10ms(100);                         // Wait for Human Ready
kbdhit=0;
while(kbdhit < 2)						// Wait for Key Release
	;
while(kbdhit >=2)
	{
	current_value = 1+ (kbdhit%4)/2; 
	acm0802_text_write(conf_messages[current_value + 8],2);
	}
eeprom_write(EEPROM_BTL_AMP_FLAG,current_value);
acm0802_text_write(conf_messages[4],1);
delay_10ms(150);                         // Wait for Human Ready

acm0802_text_write(conf_messages[5],1);	// Ask Startup Mode?
acm0802_text_write(conf_messages[6],2);
kbdhit=0;
while(kbdhit < 2)					// Wait for Key Release
	;
while(kbdhit >=2)
	{
	current_value = (kbdhit%8)/2; 
	acm0802_text_write(mode_text[current_value],2);
	}
eeprom_write(EEPROM_DEFAULT_DISP,current_value);
acm0802_text_write(conf_messages[4],1);
delay_10ms(150);                         // Wait for Human Ready
acm0802_text_write(conf_messages[0],1);
acm0802_text_write(conf_messages[7],2);
delay_10ms(300);                         // Wait for Human Ready
}

/*
----< Opening Splash Message >----
*/
void	opening_message(void)
{
char	*cpoint,length,loop;
cpoint = ad_message;
length = strlen(ad_message);

	acm0802_text_write("CMM-1000",1);
loop =0;
do	{
	strncpy(sbuf,(ad_message+loop),8);
	acm0802_text_write(sbuf,2);
	delay_10ms(15);  	
	}	
while(++loop <= length-8);	
}

/*
====<< Main >>====
*/
unsigned char loop;
void main()
{
unsigned int	adc_right, adc_big;

  ANSEL       = 0x0d;              // Configure AN0,AN2,AN3,AN5 pin as analog
  ANSELH      =    0;              // Configure  pin as Digital
  TRISA       = 0xFF;              // PORTA AN0,AN2,AN3,AN5 are input
  PORTB       = 0xff;
  TRISB       = 0x10;                    // Set RB4 as Input others are output
  PORTC       = 0xff;
  TRISC       =    0;                    // Set PORTC as output
//  WPUA        = 0x00;              // No Weak Pullup for port A
  OPTION	  = 0b10000000;
  digit       =    0;
  hold_timer = HOLDTIME;
  hold_value = 0;

  __delay_ms(50);                        // Wait for LCD Ready
  ADCON1=0x60;			// A/D Clock Fosc/64 == 15kHz
  ADCON0=0b10000001;	// A2D ͂ON, Channel 0 (RA0),Righft justified
  kbdhit =0;			// Initalize ADC block
  INTCON  = 0xA0;       				// Enable GIE, T0IE
  acm0802_init();    					// Initailize LCD Module
  CustomChar();							// Load Fonts
  acm0802_reg_write(LCD_DISPLAY_SHIFT_LEFT);
  delay_10ms(20);  	
  if(kbdhit >= 2)	// If key Holded
	{
	configmenu();	// Setup EEPROM for Default Parameters
	}
  else	  opening_message();

	kbdhit = 0;
	mode = eeprom_read(EEPROM_DEFAULT_DISP)%4;		// Recall Initial Display mode
	spkimp = eeprom_read(EEPROM_SPEAKER_IMP)%16;	// Recall Recall load setting
	amptype = eeprom_read(EEPROM_BTL_AMP_FLAG)%3;	// Recall 2 when BTL, Conventional is 1
    acm0802_text_write(mode_text[mode], 1);

   do {
      if(kbdhit) //Stop if Pushed
             {
             kbdhit = 0;
             mode= (mode +1)%4;
        	 acm0802_text_write(mode_text[mode], 1);
			 if( mode == 3)
				{
				acm0802_text_write(conf_messages[11], 2);
				delay_10ms(200);                        // Wait for LCD Ready
				}
             }
      switch(mode)
             {
             case 0:
                 Disp_Voltage(ADC_Read(1));
//					Disp_Digits(ADC_Read(0),0xa5);	// Read Trimmer ADC
                  break;

             case 1:
				  adc_big = ADC_Read(3);
				  if( adc_big < (adc_right =ADC_Read(2)))
					adc_big = adc_right;
                  Disp_Power(adc_big);
                  break;

             case 2:
				  adc_big = ADC_Read(3);
				  if( adc_big < (adc_right =ADC_Read(2)))
					adc_big = adc_right;
                  hr_ref = ADC_Read(0);
                  Disp_Headroom(adc_big);
                  break;

            case 3:
//                  Disp_Ref(ADC_Read(2));
                  hr_ref = ADC_Read(0);
				  Disp_LevelBar(ADC_Read(3),1);
				  Disp_LevelBar(ADC_Read(2),2);
                  break;
             }
  __delay_ms(30);                        // Wait for LCD Ready
  } while(1);      // endless loop
}

